/********************************************************************\

  Central data collector of the Evidence Control System
  
  - DColl subscribes to all services given by the configuration
    server and writes these to the data file at every update.
  - One data file per day is generated.
  - A history of events is kept within a ring buffer for each service. Each
	entry will be the result of a conversion to double of the text 
	written to the data file. Only if the new value has changed by a
	minimum amout it will be added to the ring buffer. The history is
	available via an rpc call.
  - The history buffers are written to disk at program termination and
    are tired to be read when adding a service.
  - The command 'DColl/Log' writes the associated string to the log file
  
  Oliver Grimm, May 2010

\********************************************************************/

#define SERVER_NAME "DColl"

#define MIN_HIST_SIZE 1024	// Minimum history buffer in bytes (> 3*sizeof(int) !)
#define LOG_FILENAME "Evidence.log"

#include "Evidence.h"

#include <string>
#include <sstream>
#include <vector>
#include <iomanip>

#include <math.h>
#include <float.h>
#include <sys/stat.h>
#include <ctype.h>
#include <sys/types.h>
#include <regex.h>

using namespace std;

//
// Class declaration
//
class DataHandler:	public DimRpc, public DimClient, public DimBrowser,
					public EvidenceServer {

	struct Item {
	  DimStampedInfo *DataItem;
	  char *Buffer;
	  unsigned int HistSize;
	  int Next;
	  double LastValue;
	  double MinAbsChange;
	};
	vector<struct Item> List;
	
	DimCommand *LogCommand;
		
    FILE *DataFile;
    FILE *LogFile;
	char *Filename;
	float DataSizeMB, LogSizeMB;
	int DataSizeLastUpdate, LogSizeLastUpdate;
	char *BaseDir;
    DimService *LogSizeService, *DataSizeService, *DataFilename;
	int SizeUpdateDelay;
	int TimeForNextFile;
	int RollOver;

	int RegExCount;
	regex_t *RegEx;
	double *MinChange;
    unsigned int *HistSize;
	
    void infoHandler();
    void rpcHandler();
    void commandHandler();
	void AddService(string);
	void RemoveService(string);
	off_t FileSize(FILE *);
	FILE *OpenHistFile(string, const char *);
   
  public:
    DataHandler();
    ~DataHandler();
}; 

//
// Constructor
//
DataHandler::DataHandler(): DimRpc("ServiceHistory", "C", "C"), EvidenceServer(SERVER_NAME) {

  // Initialization to prevent freeing unallocated memory 
  DataFile = NULL;
  LogFile = NULL;
  Filename = NULL;
  
  LogSizeService = NULL;
  DataSizeService = NULL;
  
  DataSizeLastUpdate = 0;
  LogSizeLastUpdate = 0;
  TimeForNextFile = 0;

  // Request configuration data
  BaseDir = GetConfig("basedir");
  SizeUpdateDelay = atoi(GetConfig("sizeupdate"));
  RollOver = atoi(GetConfig("rollover"));

  // Open log file
  if ((LogFile = fopen((string(BaseDir)+"/"+LOG_FILENAME).c_str(), "a")) == NULL) {
    Message(FATAL, "Could not open log file (%s)", strerror(errno));
  }

  // Provide logging command   
  LogCommand = new DimCommand("DColl/Log", (char *) "C", this);
              
  // Create services for file sizes and data file name
  DataSizeMB = 0;
  DataSizeService = new DimService(SERVER_NAME "/DataSizeMB", DataSizeMB);
  
  LogSizeMB = FileSize(LogFile)/1024.0/1024.0;
  LogSizeService = new DimService(SERVER_NAME "/LogSizekB", LogSizeMB);

  DataFilename = new DimService(SERVER_NAME "/CurrentFile", (char *) "");

  // Count how many minimum change regular expressions are present
  char *Change = GetConfig("items");
  RegExCount = 0;
  char *Token = strtok(Change, "\t ");
  while (Token != NULL) {
	RegExCount++;
	Token = strtok(NULL, "\t ");
  }
  
  // Allocate memory for regular expressions, minimum change and history size
  RegEx = new regex_t [RegExCount];
  MinChange = new double [RegExCount];
  HistSize = new unsigned int [RegExCount];

  // Compile regular expressions and extract minimum change and history size
  int Pos = 0;
  for (int i=0; i<RegExCount; i++) {
    int Len = strlen(Change+Pos) + 1;
    Token = strtok(Change + Pos, ": \t");

    int Ret = regcomp(&RegEx[i], Token, REG_EXTENDED|REG_NOSUB);
	if (Ret != 0) {
	  char ErrMsg[200];
	  regerror(Ret, &RegEx[i], ErrMsg, sizeof(ErrMsg));
	  RegExCount--;
	  i--;
	  Message(ERROR, "Error compiling regular expression '%s' (%s)", Token, ErrMsg);
	}
	else {
	  if ((Token=strtok(NULL, ": \t")) != NULL) MinChange[i] = atof(Token);
	  else MinChange[i] = 0;

  	  if ((Token=strtok(NULL, "")) != NULL) HistSize[i] = atoi(Token)*1024;
	  else HistSize[i] = 0;
	}
	Pos += Len;
  }

  // Subscribe to list of servers at DIS_DNS 
  AddService("DIS_DNS/SERVER_LIST");
}

//
// Destructor: Close files and free memory
//
DataHandler::~DataHandler() {

  // Delete all DIM subscriptions
  while (List.size() != 0) RemoveService(List[0].DataItem->getName());
  
  delete LogCommand;
  delete DataFilename;
  delete[] Filename;
  
  delete LogSizeService;
  delete DataSizeService;

  // Close files
  if (LogFile != NULL) if (fclose(LogFile) != 0) {
	Message(ERROR, "Could not close log file (%s)", strerror(errno));
  }
  if (DataFile != NULL && fclose(DataFile) != 0) {
	Message(ERROR, "Error: Could not close data file (%s)", strerror(errno));
  }

  // Free memory for regular expressions handling
  for (int i=0; i<RegExCount; i++) {
    regfree(&RegEx[i]);
  }
  delete[] MinChange;
  delete[] RegEx;
}

//
// Implementation of data handling
//
// DIM ensures infoHandler() is called serialized, therefore
// no mutex is needed to serialize writing to the file
void DataHandler::infoHandler() {

  DimInfo *Info = getInfo();

  // Check if service available
  if (!ServiceOK(Info)) return;

  //
  // ====== Part A: Handle service subscriptions ===
  //
  // Services are added here, removal only in constructor.
  
  // If service is DIS_DNS/SERVER_LIST, subscribe to all SERVICE_LIST services
  if (strcmp(Info->getName(), "DIS_DNS/SERVER_LIST") == 0) {	
	char *Token = strtok(Info->getString(), "+-!@");	
	while (Token != NULL) {
	  AddService(string(Token)+"/SERVICE_LIST"); // 'add' also for '-' and '!'
	  Token = strtok(NULL, "|"); // Skip server IP address
	  Token = strtok(NULL, "@");
	}	
	return;
  }

  // If service is SERVICE_LIST of any server, scan all services.
  // Subscribe to all services (but not to commands and RPCs)
  if (strstr(Info->getName(), "/SERVICE_LIST") != NULL) {
	char *Name = strtok(Info->getString(), "+-!|");
	while (Name != NULL) {
	  // Check if item is a service
	  char *Type = strtok(NULL, "\n");
	  if (Type == NULL) return; // for safety, should not happen
      if (strstr(Type, "|CMD")==NULL && strstr(Type, "|RPC")==NULL) AddService(Name); // 'add' also for '-' and '!'
	  Name = strtok(NULL, "|");
	}
	return;
  }

  //
  // ====== Part B: Handle opening data files ===
  //

  // If it is time to open new data file, close the current one 
  if (time(NULL) >= TimeForNextFile) {
	if (DataFile != NULL && fclose(DataFile) != 0) {
	  Message(ERROR, "Error: Could not close data file (%s)", strerror(errno));
	}
	DataFile = NULL;
  }

  // Open new data file if necessary
  if (DataFile == NULL) {
	time_t Time = time(NULL);
	struct tm *T = localtime(&Time);

	// Get time structure with date rollover
	if (T->tm_hour >= RollOver) T->tm_mday++;
	if (mktime(T) == -1) Message(ERROR, "mktime() failed, check filename");

	// Create direcory if not existing (ignore error if already existing)
	char *Dir;
	if (asprintf(&Dir, "%s/%d%02d", BaseDir, T->tm_year+1900, T->tm_mon + 1) == -1) {
	  Message(FATAL, "asprintf() failed, could not create direcory name");	
	}
	if(mkdir(Dir, S_IRWXU|S_IRWXG)==-1 && errno!=EEXIST) {
	  Message(FATAL, "Could not create directory '%s' (%s)", Dir, strerror(errno));
	}

	// Create filename
	free(Filename);
	if (asprintf(&Filename, "%s/%d%02d%02d.slow", Dir, T->tm_year+1900, T->tm_mon+1, T->tm_mday) == 1) {
	  Message(FATAL, "asprintf() failed, could not create filename");		  
	}
	free(Dir);
	
	// Open file
	if ((DataFile = fopen(Filename, "a")) == NULL) {
      Message(FATAL, "Could not open data file '%s' (%s)", Filename, strerror(errno));
	}
	else Message(INFO, "Opened data file '%s'", Filename);
	DataFilename->updateService(Filename);
	
	// Calculate time for next file opening
	T->tm_sec = 0;
	T->tm_min = 0;
	T->tm_hour = RollOver;
	TimeForNextFile = mktime(T);
  }

  //
  // ====== Part C: Handle writing to data file ===
  //

  // Identify index of service
  int Service;
  for (Service=0; Service<List.size(); Service++) if (Info == List[Service].DataItem) break;
  if (Service == List.size()) return;  // Service not found

  // If negative value for absolute change, do not write to file
  if (List[Service].MinAbsChange >= 0) {
	// Write data header
	time_t RawTime = Info->getTimestamp();
	struct tm *TM = localtime(&RawTime);

	fprintf(DataFile, "%s %d %d %d %d %d %d %d %lu ", Info->getName(), TM->tm_year+1900, TM->tm_mon+1, TM->tm_mday, TM->tm_hour, TM->tm_min, TM->tm_sec, Info->getTimestampMillisecs(), Info->getTimestamp());

	// Translate data into ASCII
	char *Text = EvidenceServer::ToString(Info);

	if (Text != NULL) {
	  // Replace new line by '\' and all other control characters by white space
	  for (int i=0; i<strlen(Text); i++) {
		if (Text[i] == '\n') Text[i] = '\\';
		else if (iscntrl(Text[i])) Text[i] = ' ';
	  }

	  // Write to file
      fprintf(DataFile, "%s\n", Text);

	  free(Text);
	}
	else fprintf(DataFile, "Cannot interpret format identifier\n");

	// Terminate if error because otherwise infinite loop might result as
	// next call to this infoHandler() will try to (re-)open file
	if(ferror(DataFile)) {
      fclose(DataFile);
	  DataFile = NULL;
	  Message(FATAL, "Error writing to data file, closed file (%s)", strerror(errno));
	}

	// Update datafile size service (not every time to avoid infinite loop)
	if (time(NULL) - DataSizeLastUpdate > SizeUpdateDelay) {
      fflush(DataFile); // not continuously to reduce load

	  DataSizeMB = FileSize(DataFile)/1024.0/1024.0;
	  DataSizeService->updateService();
	  DataSizeLastUpdate = time(NULL);
	}  
  } // Check for disk writing

  //
  // ====== Part D: Handle history service ===
  //

  if (Info->getSize()==0 || Info->getTimestamp()==0) return;

  // Check if data should be added to history buffer
  if (strcmp(Info->getFormat(),"C") != 0 && strlen(Info->getFormat())==1) {
	// Calculate sum of all number in array
	char *Text = EvidenceServer::ToString(Info);
	char *Token = strtok(Text, " ");
	double Sum = 0;
	while (Token != NULL) {
	  Sum += atof(Token);
      Token = strtok(NULL, " ");
	}
	free(Text);
	// Minimum change?
	if (fabs(Sum-List[Service].LastValue) < fabs(List[Service].MinAbsChange)) return;
	List[Service].LastValue = Sum;
  }
  
  // Check if data fits into buffer
  if (List[Service].HistSize < Info->getSize() + sizeof(int)+ 2*sizeof(EvidenceHistory::Item)) return;

  int Size = Info->getSize() + 2*sizeof(EvidenceHistory::Item), Next = List[Service].Next;
  void *WrapPos = NULL;
  char *Buffer = List[Service].Buffer;
  int Oldest = *(int *) Buffer;

  // Check if buffer wrap-around (write wrap mark after Oldest is adjusted)
  if (Next + Size >= List[Service].HistSize) {
    WrapPos = Buffer + Next;
    Next = 4;
  }

  // Adapt pointer to oldest entry
  while ((Oldest < Next + Size) && 
  		 (Oldest + *((int *) (Buffer + Oldest) + 1) + 2*sizeof(int) > Next)) {
	// Check for wrap-around
	if (memcmp(Buffer + Oldest, &EvidenceHistory::WrapMark, sizeof(EvidenceHistory::WrapMark)) == 0) {
	  Oldest = 4;
	  continue;
	}
	// Check if end marker reached, then only one event fits buffer
  	if (memcmp(Buffer + Oldest, &EvidenceHistory::EndMark, sizeof(EvidenceHistory::EndMark)) == 0) {
	  Oldest = Next;
	  break;
	}
	// Move to next entry
    Oldest += *((int *) (Buffer + Oldest) + 1) + 2*sizeof(int);
  }
  // Update pointer in buffer
  *(int *) Buffer = Oldest;

  // Write wrap mark if necessary
  if (WrapPos != NULL) memcpy(WrapPos, &EvidenceHistory::WrapMark, sizeof(EvidenceHistory::WrapMark));

  // Copy data into ring buffer
  *((int *) (Buffer + Next)) = Info->getTimestamp();
  *((int *) (Buffer + Next + sizeof(int))) = Info->getSize();
  memcpy(Buffer + Next + 2*sizeof(int), Info->getData(), Info->getSize());

  // Adjust pointer for next entry and write end marker to buffer
  Next += Info->getSize() + sizeof(EvidenceHistory::Item);
  memcpy(Buffer + Next, &EvidenceHistory::EndMark, sizeof(EvidenceHistory::EndMark));
  
  List[Service].Next = Next;
}

//
// Implementation of log writing
//
void DataHandler::commandHandler() {
 
  if (getCommand() != LogCommand || LogFile == NULL) return;

  // Replace all carriage-return by line feed and non-printable characters
  char *Text = getCommand()->getString();
  for (unsigned int i=0; i<strlen(Text); i++) {
    if (Text[i] == '\r') Text[i] = '\n';
	if(isprint(Text[i])==0 && isspace(Text[i])==0) Text[i] = ' ';
  }
 
  time_t RawTime = time(NULL);
  struct tm *TM = localtime(&RawTime);

  fprintf(LogFile, "%2d/%2d/%4d %2d:%2d:%2d %s (ID %d): %s\n",
  		TM->tm_mday, TM->tm_mon+1, TM->tm_year+1900,
		TM->tm_hour, TM->tm_min, TM->tm_sec, getClientName(), getClientId(), Text);

  fflush(LogFile);
  
  // If error close file (otherwise infinite loop because Message() also writes to log) 
  if(ferror(LogFile)) {
    fclose(LogFile);
	LogFile = NULL;
    Message(ERROR, "Error writing to log file, closing file (%s)", strerror(errno));
  }
    
  // Update logfile size service
  if (time(NULL) - LogSizeLastUpdate > SizeUpdateDelay) {
	LogSizeMB = FileSize(LogFile)/1024.0/1024.0;
	LogSizeService->updateService();
	LogSizeLastUpdate = time(NULL);
  }
}


//
// Implementation of history buffer distribution
//
void DataHandler::rpcHandler() {

  // Search for history buffer
  for (int i=0; i<List.size(); i++) {
    if (strcmp(List[i].DataItem->getName(), getString()) == 0) {
	  setData((void *) List[i].Buffer, List[i].HistSize);
	  return;
	}
  }

  // Try to open history file if not found in memory
  FILE *File = OpenHistFile(getString(), "rb");
  if (File == NULL) {
    setData(NULL, 0);
	return;
  }

  // Read history file
  off_t Size = FileSize(File);
  if (Size != -1) {
	char *Buffer = new char [Size-sizeof(int)];
	fseek(File, sizeof(int), SEEK_SET);
	fread(Buffer, sizeof(char), Size-sizeof(int), File);
	if (ferror(File) != 0) {
	  Message(WARN, "Error reading history file '%s' in rpcHandler()", getString());
	  setData(NULL, 0);		// Default response
	}
	else setData((void *) Buffer, Size);
	delete[] Buffer;
  }
  
  if (fclose(File) != 0) Message(WARN, "Error closing history file '%s' in rpcHandler()", getString());
}


//
// Add service to watch list
//
void DataHandler::AddService(string Name) {

  // Check if already subscribed to this service
  for (int i=0; i<List.size(); i++) {
	if (Name == List[i].DataItem->getName()) return;
  }

  // Set minimum required change by comparing to regular expressions
  struct Item New;
  New.MinAbsChange = 0;
  New.HistSize = 0;
  for (int i=0; i<RegExCount; i++) {
    if (regexec(&RegEx[i], Name.c_str(), (size_t) 0, NULL, 0) == 0) {
	  New.MinAbsChange = MinChange[i];
	  if (HistSize[i] != 0) New.HistSize = HistSize[i];
	}
  }

  // At least 3*sizeof(int)
  if (New.HistSize < MIN_HIST_SIZE) New.HistSize = MIN_HIST_SIZE;

  // Create history service
  New.Buffer = new char [New.HistSize];
  memset(New.Buffer, 0, New.HistSize);
  *(int *) New.Buffer = 4;
  New.Next = 4;
  New.LastValue = DBL_MAX;

  // Load history buffer from file if existing
  FILE *File = OpenHistFile(Name, "rb");
  if (File != NULL) {
    // Only load if current buffer size is equal or larger
    if (FileSize(File) <= New.HistSize*sizeof(char)+sizeof(New.Next) && FileSize(File) != -1) {
      fread(&New.Next, sizeof(New.Next), 1, File);
      fread(New.Buffer, sizeof(char), New.HistSize, File);
	  if (ferror(File) != 0) Message(WARN, "Error reading history file '%s' in AddService()", Name.c_str());
      if (fclose(File) != 0) Message(WARN, "Error closing history file '%s' in AddService()", Name.c_str());;
	}
  }

  // Subscribe to service
  New.DataItem = new DimStampedInfo(Name.c_str(), NO_LINK, this);

  // Add item to list
  List.push_back(New);
}


//
// Remove service from watch list
//
void DataHandler::RemoveService(string Name) {

  // Find service index
  vector<struct Item>::iterator E;
  for (E=List.begin(); E<List.end(); ++E) if (Name == (*E).DataItem->getName()) {
	// Delete subscription first so handler and not called anymore
	delete (*E).DataItem;

	// Save history buffer
	FILE *File = OpenHistFile(Name, "wb");
	if (File != NULL) {
      fwrite(&(*E).Next, sizeof((*E).Next), 1, File);
      fwrite((*E).Buffer, sizeof(char), (*E).HistSize, File);
	  if (ferror(File) != 0) Message(WARN, "Error writing history file '%s' in RemoveService()", Name.c_str());
      if (fclose(File) != 0) Message(WARN, "Error closing history file '%s' in RemoveService()", Name.c_str());;
	}
	
	// Delete history service and free memory
    delete[] (*E).Buffer;
	List.erase(E);
  }
}

//
// Determine size of file in kB
//
off_t DataHandler::FileSize(FILE *File) {

  struct stat FileStatus;

  if (fstat(fileno(File), &FileStatus) == -1) {
     Message(WARN, "Could not determine size of file (%s)", strerror(errno));
	 return -1;
  }

  return FileStatus.st_size;
}

//
// Open file for service history
//
FILE *DataHandler::OpenHistFile(string Service, const char *Mode) {

  string Dir = string(BaseDir) + "/Histories/";

  // Create directory if not yet existing
  if(mkdir(Dir.c_str(), S_IRWXU|S_IRWXG)==-1 && errno!=EEXIST) return NULL;

  // Replace all '/' by '_' in string and open file
  for (int i=0; i<Service.size(); i++) if (Service[i] == '/') Service[i] = '_';
  return fopen((Dir + Service).c_str(), Mode);
}

//	    
// Main program
//
int main() {
        
  // Static ensures calling of destructor by exit()
  static DataHandler Data;
  
  // Sleep until signal caught
  pause();
}
